from copy import deepcopy
import math
from random import uniform
from typing import List
import time
from mlassistant.core import Model, ModelIO
import torch
from torch.nn import functional as F
from torch import nn
from ......utils.losses.balanced_focal_cross_entropy import balanced_focal_cross_entropy_loss_semi
from ......utils.generalized_dice import dice_loss
from ......utils.losses.ent_losses import entropy_loss_normalized
from ......utils.losses.tverskyLoss import tversky_loss

from ......utils.dynamic_num_generator.num_generator import init_num_generator1, init_num_generator2
from ......utils.dynamic_num_generator.wrapper import dynamic_num_generator





class RESUNET(Model):
    def __init__(self,
            base_model: Model,
            gamma = 2 , smooth=1, alpha=0.7, beta=0.3 , ent_factor = lambda x : 0.1) -> None:
        super(RESUNET, self).__init__()

        
        self.base = base_model
        self.gamma = gamma
        self.smooth = smooth
        self.alpha = alpha
        self.beta = beta
        self.ent_factor = ent_factor
        print(self.gamma)
        self.epochNUM = 1
        print(self.epochNUM)
    

    def train(self,mode : bool=True):
        super().train(mode)
        self.epochNUM += 1

    def forward(self,
            mammo_x: torch.Tensor,
            mammo_loss_and_gt: torch.Tensor) -> ModelIO:
        """first forwards supervised samples through student net, then all samples through both nets. """
        output = dict(loss=0.0)
        assert mammo_loss_and_gt.shape[1] == 2 , "loss and gt is not in expected shape (B,2,size,size)"
        loss_type = mammo_loss_and_gt[:,0,:,:] 
        mask = mammo_loss_and_gt[:,1,:,:]
        network_output = self.base.forward(mammo_x).softmax(dim=1) # logsoftmax # B 
        network_output_soft = network_output #torch.exp(network_output).to(mammo_x.device) # softmax output of model
        output['pixel_probs'] = network_output_soft

        main_shape = mask.shape
        network_output_shape = network_output.shape



        loss_channels = loss_type.unsqueeze(dim=1).expand(main_shape[0],network_output_shape[1],main_shape[1], main_shape[2])
        mask = mask.unsqueeze(dim = 1)
        mask_channels = torch.cat((1-mask, mask), 1)

        dummy_mask = mask_channels.clone().to(mammo_x.device)
        dummy_net_out = network_output_soft.clone().to(mammo_x.device)
        dummy_mask[loss_channels == 0] = 0.0
        dummy_shape = dummy_mask.shape
        dummy_mask = torch.cat((dummy_mask,torch.ones((dummy_shape[0],1,dummy_shape[2],dummy_shape[2])).to(mammo_x.device)), 1)
        dummy_net_out = torch.cat((dummy_net_out,torch.ones((dummy_shape[0],1,dummy_shape[2],dummy_shape[2])).to(mammo_x.device)), 1)
        # dummy_yhat = network_output_soft * loss_channels
        # dummy_y = mask_channels * loss_channels
        focal_loss = balanced_focal_cross_entropy_loss_semi(
                dummy_net_out,
                dummy_mask,
                focal_gamma=self.gamma)
        #dice = dice_loss(dummy_net_out,dummy_mask)
        #tversky = tversky_loss(inputs = dummy_yhat[:,1,...], targets = dummy_y[:,1,...], smooth=self.smooth, alpha = self.alpha,  beta= self.beta)
        #ce_loss = F.binary_cross_entropy(dummy_yhat, dummy_y)
        
        entropy_loss_val = entropy_loss_normalized(network_output_soft)
        entropy_suploss_val = torch.mean(entropy_loss_val * loss_type)
        entropy_unsuploss_val = torch.mean(entropy_loss_val * (1-loss_type))
        sup_ratio = torch.sum(loss_type == 1)
        all_cel = torch.prod(torch.tensor(loss_type.shape)).to(mammo_x.device)
        entropy_sup = entropy_suploss_val * (all_cel/sup_ratio)
        entropy_unsup = entropy_unsuploss_val * (all_cel/(all_cel - sup_ratio + 1e-30))
        entropy_loss_val = torch.mean(entropy_loss_val)
        entloss = torch.mean(torch.tensor([entropy_sup, entropy_unsup])).to(mammo_x.device)

        coef = self.ent_factor(int(self.epochNUM/2))
        
        #output['ce_loss'] = ce_loss
        output['loss'] = focal_loss + coef * entropy_loss_val
        output['totalloss'] = output['loss']
        #output['diceloss'] = dice
        output['focalloss'] = focal_loss
        output['suploss'] = focal_loss
        output['unsuploss'] = entropy_loss_val
        output['coef_loss'] = torch.tensor(coef).to(mammo_x.device)
        #output['tverskyloss'] = tversky
        output['org_pixel_labels'] = mammo_loss_and_gt
        output["entallloss"] = entropy_loss_val
        output["supentloss"] = entropy_sup
        output["unsupentloss"] = entropy_unsup
        output["entloss"] =  entloss
        return output 